"""Константы."""
import logging


# максимальная длина сообщения в байтах
MAX_PACKAGE_LENGTH = 10240

# кодировка проекта
ENCODING = 'utf-8'

# максимальная очередь подключений
MAX_CONNECTIONS = 5

# параметры подключения к серверу по умолчанию
DEFAULT_ADDRESS = '127.0.0.1'
DEFAULT_PORT = 7777

# Текущий уровень логирования
LOGGING_LEVEL = logging.DEBUG
